/*
    kdtree.c - locational code construction and handling.
    
    Copyright (C) 2009  Laurent Poirrier
    This file is part of the lin-kd-tree library.

    The lin-kd-tree library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, version 3 of the License.

    The lin-kd-tree library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <string.h>
#include "kd-tree.h"

#define DEBUG(s...)	fprintf(stderr, s)
#define DEBUG_VAR	DEBUG


// ***********************************************************************
// Order: (a < b) or ((a == b) && (a.depth < b.depth))
// ***********************************************************************
#if KDTREE_SIMPLE_PATH
int  kdtree_before(const kdtree_node *a, const kdtree_node *b)
{
	return( KDTREE_BEFORE(*a, *b) );
}

#else

int  kdtree_before(const kdtree_node *a, const kdtree_node *b)
{
	int e;

	// most significant element first
	// bits after the (depth) first bits must be zero
	for (e = 0; e < KDTREE_PATH_ELEMS; e++) {
		if (a->path[e] < b->path[e]) return(1);
		if (a->path[e] > b->path[e]) return(0);
	}
	
	// if a is prefix of b (b's ext. being 0s), then a is before
	if (a->depth < b->depth)
		return(1);
	
	// if b is prefix of a (idem), or b equals a, then a is not before
	return(0);
}

#endif

// ***********************************************************************
// Intersecton: (a is prefix of b, or equal) or (b is prefix of a)
// ***********************************************************************
#if KDTREE_SIMPLE_PATH

int  kdtree_intersect(const kdtree_node *a, const kdtree_node *b)
{
	return( KDTREE_INTERSECT(*a, *b) );
}

#else 

int  kdtree_intersect(const kdtree_node *a, const kdtree_node *b)
{
	int l, e, el;
	
	// the first l bits must be equal
	l = KDTREE_MIN_DEPTH(*a, *b);
	
	// Compare all bits of the first elements
	el = l >> KDTREE_PATH_ELEM_LOG2;
	for (e = 0; e < el; e++) {
		if (a->path[e] != b->path[e]) return(0);
	}
	
	// Compare the last bits
	l -= el << KDTREE_PATH_ELEM_LOG2;
	if (l == 0) return(1);
	
	if ((a->path[el] >> (KDTREE_PATH_ELEM_BITS - l))
	!=  (b->path[el] >> (KDTREE_PATH_ELEM_BITS - l)))
		return(0);
	
	return(1);
}

#endif

// ***********************************************************************
// Ordered intersection: (a is prefix of b or equal)
// ***********************************************************************
#if KDTREE_SIMPLE_PATH
int  kdtree_ordered_intersect(const kdtree_node *a, const kdtree_node *b)
{
	return( KDTREE_ORDERED_INTERSECT(*a, *b) );	
}

#else
int  kdtree_ordered_intersect(const kdtree_node *a, const kdtree_node *b)
{
	int l, e, el;
	
	if (b->depth < a->depth) return(0);
	
	// Compare all bits of the first elements
	el = a->depth >> KDTREE_PATH_ELEM_LOG2;
	for (e = 0; e < el; e++) {
		if (a->path[e] != b->path[e]) return(0);
	}
	
	// Compare the last bits
	l = a->depth - (el << KDTREE_PATH_ELEM_LOG2);
	if (l == 0) return(1);
	
	if ((a->path[el] >> (KDTREE_PATH_ELEM_BITS - l))
	!=  (b->path[el] >> (KDTREE_PATH_ELEM_BITS - l)))
		return(0);
	
	return(1);
}

#endif

// ***********************************************************************
// 
// ***********************************************************************
#if ((KDTREE_SIMPLE_PATH) \
  && ((KDTREE_COORD_BITS == 32) || (KDTREE_COORD_BITS == 64)) \
  && ((KDTREE_PATH_ELEM_BITS == 32) || (KDTREE_PATH_ELEM_BITS == 64)))

#if (KDTREE_PATH_ELEM_BITS == 32)

// 01234567 89a***** ******** ********
// 01234567 -------- -------- 89a-----
// 0123---- ----4567 -------- 89a-----
// 01----23 ----45-- --67---- 89----a-
// 0--1--2- -3--4--5 --6--7-- 8--9--a-
static uint32_t explode(KDTREE_COORD n)
{
#if (KDTREE_COORD_BITS == 64)
	n >>= 32;
#endif
	n &= 0xffe00000;
	n = (n | (n >> 16)) & 0xff0000e0;
	n = (n | (n >>  8)) & 0xf00f00e0;
	n = (n | (n >>  4)) & 0xc30c30c2;
	n = (n | (n >>  2)) & 0x92492492;
	return(n);
}


#elif (KDTREE_PATH_ELEM_BITS == 64)

// 01234567 89abcdef ghijkl** ******** ******** ******** ******** ********
// 01234567 89abcdef -------- -------- -------- -------- ghijkl-- --------
// 01234567 -------- -------- 89abcdef -------- -------- ghijkl-- --------
// 0123---- ----4567 -------- 89ab---- ----cdef -------- ghij---- ----kl--
// 01----23 ----45-- --67---- 89----ab ----cd-- --ef---- gh----ij ----kl--
// 0--1--2- -3--4--5 --6--7-- 8--9--a- -b--c--d --e--f-- g--h--i- -j--k--l
static uint64_t explode(uint64_t n)
{
#if (KDTREE_COORD_BITS == 32)
	n = (n >> 10) << 42;
#else
	n &= 0xfffffc0000000000;
#endif
	n = (n | (n >> 32)) & 0xffff00000000fc00;
	n = (n | (n >> 16)) & 0xff0000ff0000fc00;
	n = (n | (n >>  8)) & 0xf00f00f00f00f00c;
	n = (n | (n >>  4)) & 0xc30c30c30c30c30c;
	n = (n | (n >>  2)) & 0x9249249249249249;
	return(n);
}
#endif

// first_bit() returns the index of the first non-0 bit, numbering
// from the most significant (index 0) to the least significant one.
#if (KDTREE_PATH_ELEM_BITS == 32) \
	&& ((defined(__i386__)) || (defined(__x86_64__)))

static uint32_t first_bit(uint32_t n)
{
	uint32_t bit;
	
	if (!n) return(32);
	__asm__ ("bsrl %1, %0\n\t" : "=r" (bit) : "rm" (n));
	return(32 - bit - 1);
}

#elif (KDTREE_PATH_ELEM_BITS == 64) && (defined(__x86_64__))

static uint64_t first_bit(uint64_t n)
{
	uint64_t bit;
	
	if (!n) return(64);
	__asm__ ("bsrq %1, %0\n\t" : "=r" (bit) : "rm" (n));
	return(64 - bit - 1);
}

#else

static int first_bit(KDTREE_PATH_ELEM_TYPE n)
{
	int depth = 0;

	if (!n) return(KDTREE_PATH_ELEM_BITS);

	while (!(n & KDTREE_PATH_ELEM_MSB)) {
		depth++;
		n <<= 1;
	}
	
	return(depth);
}

#endif

void kdtree_place(kdtree_node *dst, const kdtree_iaabb *bb)
{
	KDTREE_PATH_ELEM_TYPE p0, p1;

	p0 = explode(bb->x0) | (explode(bb->y0) >> 1) | (explode(bb->z0) >> 2);
	p1 = explode(bb->x1) | (explode(bb->y1) >> 1) | (explode(bb->z1) >> 2);
	
	dst->depth = first_bit(p0 ^ p1);
	
	dst->path[0] = p0 & ~(KDTREE_PATH_ELEM_ONES >> dst->depth);
}

#else

void kdtree_place(kdtree_node *dst, const kdtree_iaabb *bb)
{
	KDTREE_COORD mask;
	KDTREE_PATH_ELEM_TYPE wbit;
	int wel, dim;
	
	memset(dst->path, 0, sizeof(dst->path));
	dst->depth = 0;

	// we interleave x, y and z bits from the AABB,
	// as long as they match for bounding box's p0 and p1.	
	/*{
		// if we are given a pre-placed bounding box and only try to
		// deepen the node:
		int skip;
		
		wel = dst->depth >> KDTREE_PATH_ELEM_LOG2;
		skip = dst->depth - (wel << KDTREE_PATH_ELEM_LOG2);
	
		wbit = (KDTREE_PATH_ELEM_MSB >> skip);
		mask = (KDTREE_COORD_MSB >> dst->depth);
	}*/
	wel = 0;
	wbit = KDTREE_PATH_ELEM_MSB;
	mask = KDTREE_COORD_MSB;
	
	dim = 0;
	while (dst->depth < KDTREE_PATH_ELEM_BITS * KDTREE_PATH_ELEMS) {
		
		switch(dim) {
		case 0:
			if ((bb->x0 ^ bb->x1) & mask) return;
			if (bb->x0 & mask) dst->path[wel] |= wbit;
			dim = 1;
			break;
		case 1:
			if ((bb->y0 ^ bb->y1) & mask) return;
			if (bb->y0 & mask) dst->path[wel] |= wbit;
			dim = 2;
			break;
		case 2:
			if ((bb->z0 ^ bb->z1) & mask) return;
			if (bb->z0 & mask) dst->path[wel] |= wbit;
			dim = 0;
			mask >>= 1;
			break;
		}
		
		wbit >>= 1;
		if (wbit == 0) {
			wbit = KDTREE_PATH_ELEM_MSB;
			wel++;
		}
		
		dst->depth++;
	}
}

#endif

// ***********************************************************************
// Division by 3
// ***********************************************************************
/*
 Using a lookup table for div3/mod3 makes the overall collision detection
 about 0.5% faster. But this relies on CPU/memory speed ratio.
 Using stdlib's div() makes it about 5% slower than dedicated '/'
 and '% 'operators.
 
 gcc detects that we want the integer result and the remainder
 of the same division by 3. It even replaces it (gcc 4.3.0 on x86_64) by
 a fixed-point multiplication by 1/3. Scary.
*/

#if ((0) && ((KDTREE_PATH_ELEMS * KDTREE_PATH_ELEM_BITS) <= 64))

static uint32_t divmod3[65][2] = {
{0, 0}, {0, 1}, {0, 2}, {1, 0}, {1, 1}, {1, 2}, {2, 0}, {2, 1}, {2, 2}, {3, 0},
{3, 1}, {3, 2}, {4, 0}, {4, 1}, {4, 2}, {5, 0}, {5, 1}, {5, 2}, {6, 0}, {6, 1},
{6, 2}, {7, 0}, {7, 1}, {7, 2}, {8, 0}, {8, 1}, {8, 2}, {9, 0}, {9, 1}, {9, 2},
{10, 0}, {10, 1}, {10, 2}, {11, 0}, {11, 1}, {11, 2}, {12, 0}, {12, 1},
{12, 2}, {13, 0}, {13, 1}, {13, 2}, {14, 0}, {14, 1}, {14, 2}, {15, 0},
{15, 1}, {15, 2}, {16, 0}, {16, 1}, {16, 2}, {17, 0}, {17, 1}, {17, 2},
{18, 0}, {18, 1}, {18, 2}, {19, 0}, {19, 1}, {19, 2}, {20, 0}, {20, 1},
{20, 2}, {21, 0}, {21, 1} };

//#define DIV3(v)	divmod3[v][0]
//#define MOD3(v)	divmod3[v][1]

#define DIVMOD3(q, r, v)	q = divmod3[v][0]; r = divmod3[v][1]

#else

//#define DIVMOD3(q, r, v)	q = ((v) * 21846) >> 16; r = (v)-(q)-(q)-(q)

//#define DIV3(v)	((v) / 3)
//#define MOD3(v)	((v) % 3)
#define DIVMOD3(q, r, v)	q = (v) / 3; r = (v) % 3

#endif

// ***********************************************************************
// 
// ***********************************************************************
void kdtree_root(kdtree_node *dst)
{
	memset(dst->path, 0, sizeof(dst->path));
	dst->depth = 0;
}


// ***********************************************************************
// 
// ***********************************************************************
int  kdtree_split_place(kdtree_node *dst8, const kdtree_iaabb *bb,
	int split_x, int split_y, int split_z)
{
	KDTREE_COORD mask, mid;
	kdtree_iaabb sbb[2];
	kdtree_node bk;
	int n0, n1, dim, depthdiv3;
	
	/*
	// Splitting disabled
	kdtree_place(dst8, bb);
	return(1);
	*/

	// place bb into the kd-tree
	kdtree_place(&bk, bb);
	
	// if we are at maximal depth, return bk
	if (bk.depth == KDTREE_PATH_ELEM_BITS * KDTREE_PATH_ELEMS) {
		memcpy(dst8, &bk, sizeof(kdtree_node));
		return(1);
	}
	
	// select the common bits + one
	//  depthdiv3 = bk.depth / 3;
	//  dim = bk.depth % 3;
	DIVMOD3(depthdiv3, dim, bk.depth);
	mask = ~(KDTREE_COORD_ONES >> (depthdiv3 + 1));

	memcpy(&sbb[0], bb, sizeof(kdtree_iaabb));
	memcpy(&sbb[1], bb, sizeof(kdtree_iaabb));

	if ((dim == 0) && (split_x)) {
		split_x--;
		mid = bb->x1 & mask;
		sbb[0].x1 = mid - 1;
		sbb[1].x0 = mid;
	} else if ((dim == 1) && (split_y)) {
		split_y--;
		mid = bb->y1 & mask;
		sbb[0].y1 = mid - 1;
		sbb[1].y0 = mid;
	} else if ((dim == 2) && (split_z)) {
		split_z--;
		mid = bb->z1 & mask;
		sbb[0].z1 = mid - 1;
		sbb[1].z0 = mid;
	} else {
		// nothing can be split further, return bk
		memcpy(dst8, &bk, sizeof(kdtree_node));
		return(1);
	}
	
	n0 = kdtree_split_place(dst8, &sbb[0], split_x, split_y, split_z);
	n1 = kdtree_split_place(dst8 + n0, &sbb[1], split_x, split_y, split_z);

	// bipartition of the original node
	if ((n0 + n1 == 2)
	&&  (dst8[0].depth == bk.depth + 1)
	&&  (dst8[1].depth == bk.depth + 1)) {
		memcpy(dst8, &bk, sizeof(kdtree_node));
		return(1);
	}
	
	// one of the subnodes has depth > bk.depth + 1
	// (auto-implied if more than 2 subnodes)
	return(n0 + n1);
}

// ***********************************************************************
// DEBUG
// ***********************************************************************
void kdtree_node_iaabb(const kdtree_node *k, kdtree_iaabb *dst)
{
	KDTREE_COORD x = 0, y = 0, z = 0;
	KDTREE_COORD mx, my, mz;
	KDTREE_PATH_ELEM_TYPE n;
	int el, l, dim;
	
	el = 0;
	n = KDTREE_PATH_ELEM_MSB;
	mx = my = mz = KDTREE_COORD_MSB;
	
	dim = 0;
	for (l = 0; l < k->depth; l++) {
		switch(dim) {
		case 0:
			if (k->path[el] & n) x |= mx;
			dim = 1;
			mx >>= 1;
			break;
		case 1:
			if (k->path[el] & n) y |= my;
			dim = 2;
			my >>= 1;
			break;
		case 2:
			if (k->path[el] & n) z |= mz;
			dim = 0;
			mz >>= 1;
			break;
		}

		n >>= 1;
		if (!n) { n = KDTREE_PATH_ELEM_MSB; el++; }
	}
	
	mx = (mx << 1) - 1;
	my = (my << 1) - 1;
	mz = (mz << 1) - 1;
	
	dst->x0 = x;
	dst->x1 = x | mx;
	dst->y0 = y;
	dst->y1 = y | my;
	dst->z0 = z;
	dst->z1 = z | mz;
}

// *************************************************
void kdtree_print(const char *name, const kdtree_node *a)
{
	int i, rel;
	KDTREE_PATH_ELEM_TYPE rb;
	
	rb = KDTREE_PATH_ELEM_MSB;
	rel = 0;
	DEBUG_VAR("%s = ", name);
	
	for (i = 0; i < a->depth; i++) {
		DEBUG_VAR("%c", (a->path[rel] & rb) ? '1' : '0');
		rb >>= 1;
		if (rb == 0) {
			rb = KDTREE_PATH_ELEM_MSB;
			rel++;
		}

		if ((i % 3) == 2) DEBUG_VAR(" ");
	}
	
	for (i = a->depth;
	i < KDTREE_PATH_ELEM_BITS * KDTREE_PATH_ELEMS; i++) {
		DEBUG_VAR(".");
		if ((i % 3) == 2) DEBUG_VAR(" ");
	}
	
	DEBUG_VAR("\n");
}

// *************************************************
static void kdtree_bin_pr(KDTREE_COORD c)
{
	int i;
	
	for (i = KDTREE_COORD_BITS - 1; i >= 0; i--) {
		if ((i & 7) == 7) DEBUG_VAR(" ");
		DEBUG_VAR("%c", (c & ((KDTREE_COORD)1 << i)) ? '1' : '0');
	}
}

void kdtree_bin_print(const char *name, KDTREE_COORD c)
{
	DEBUG_VAR("%s = ", name);
	kdtree_bin_pr(c);
	DEBUG_VAR("\n");
}

// *************************************************
void kdtree_iaabb_print(const char *name, const kdtree_iaabb *bb)
{
	DEBUG_VAR("%s", name);
	DEBUG_VAR("\n\t->x0 ="); kdtree_bin_pr(bb->x0);
	DEBUG_VAR("\n\t->x1 ="); kdtree_bin_pr(bb->x1);
	DEBUG_VAR("\n\t->y0 ="); kdtree_bin_pr(bb->y0);
	DEBUG_VAR("\n\t->y1 ="); kdtree_bin_pr(bb->y1);
	DEBUG_VAR("\n\t->z0 ="); kdtree_bin_pr(bb->z0);
	DEBUG_VAR("\n\t->z1 ="); kdtree_bin_pr(bb->z1);
	DEBUG_VAR("\n");
}


