#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include "genbench.h"

#define DEBUG(s...)	fprintf(stderr, s)

#define CELL_MAX	16

typedef struct cgrid_cell_t {
	int nid;
	int id[CELL_MAX];
	struct cgrid_cell_t *next;
} cgrid_cell;

typedef struct cgrid_t {
	double cell_width, inv_cell_width;
	double x0, y0, z0;
	int xcells, ycells, zcells;
	cgrid_cell **data;
	
	// stats
	int nodes;
} cgrid;

cgrid *cgrid_new()
{
	cgrid *g;
	int i;
	
	g = (cgrid *)malloc(sizeof(cgrid));
	if (!g)
		return(NULL);
	
	g->cell_width = gb_max_radius() * 1.0625;
	g->inv_cell_width = 1.0 / g->cell_width;
	
	g->x0 = gb_min_x(); g->y0 = gb_min_y(); g->z0 = gb_min_z();
	g->xcells = ceil((gb_max_x() - g->x0) * g->inv_cell_width);
	g->ycells = ceil((gb_max_y() - g->y0) * g->inv_cell_width);
	g->zcells = ceil((gb_max_z() - g->z0) * g->inv_cell_width);
	
	DEBUG("%d x %d x %d cells of width %g\n",
		g->xcells, g->ycells, g->zcells, g->cell_width);
	
	g->data = (cgrid_cell **)malloc(sizeof(cgrid_cell *)
		* g->xcells * g->ycells * g->zcells);
	if (!g->data) {
		free(g);
		return(NULL);
	}
	
	for (i = 0; i < g->xcells * g->ycells * g->zcells; i++) {
		g->data[i] = NULL;
	}

	g->nodes = 0;

	return(g);
}

void cgrid_reset(cgrid *g)
{
	int i;
	cgrid_cell *c, *c2;
	
	for (i = 0; i < g->xcells * g->ycells * g->zcells; i++) {
		c = g->data[i];
		while (c) {
			c2 = c;
			c = c->next;
			free(c2);
		}
		
		g->data[i] = NULL;
	}
	
	g->nodes = 0;
}

static int cgrid_cell_offs(cgrid *g, int cx, int cy, int cz)
{
	if ((cx < 0) || (cx >= g->xcells)
	||  (cy < 0) || (cy >= g->ycells)
	||  (cz < 0) || (cz >= g->zcells))
		return(-1);
	
	return( (((cz * g->ycells) + cy) * g->xcells) + cx );
}

void cgrid_add_sphere(cgrid *g, double x, double y, double z, int n)
{
	int cx, cy, cz;
	int offs;
	cgrid_cell *c;
	
	cx = floor((x - g->x0) * g->inv_cell_width);
	cy = floor((y - g->y0) * g->inv_cell_width);
	cz = floor((z - g->z0) * g->inv_cell_width);
	
	offs = cgrid_cell_offs(g, cx, cy, cz);
	if (offs == -1) {
		DEBUG("Object %d (%d, %d, %d) is out of bounds\n",
			n, cx, cy, cz);
		exit(1);
	}
	
	c = g->data[offs];

	if ((!c) || (c->nid >= CELL_MAX)) {
		c = malloc(sizeof(cgrid_cell));
		if (!c) {
			DEBUG("OOM\n");
			exit(1);
		}
		
		c->nid = 0;
		c->next = g->data[offs];
		g->data[offs] = c;

		g->nodes++;
	}
	
	c->id[c->nid] = n;
	c->nid++;
}

/*
...
...
...

...
.0I
III

III
III
III
*/

void cgrid_cell_interactions(cgrid *g, int cx, int cy, int cz)
{
	int offs[14];
	cgrid_cell *p, *q;
	int i, j, c;
	
	offs[ 0] = cgrid_cell_offs(g, cx    , cy    , cz);
	offs[ 1] = cgrid_cell_offs(g, cx + 1, cy    , cz);
	offs[ 2] = cgrid_cell_offs(g, cx - 1, cy + 1, cz);
	offs[ 3] = cgrid_cell_offs(g, cx    , cy + 1, cz);
	offs[ 4] = cgrid_cell_offs(g, cx + 1, cy + 1, cz);
	offs[ 5] = cgrid_cell_offs(g, cx - 1, cy - 1, cz + 1);
	offs[ 6] = cgrid_cell_offs(g, cx    , cy - 1, cz + 1);
	offs[ 7] = cgrid_cell_offs(g, cx + 1, cy - 1, cz + 1);
	offs[ 8] = cgrid_cell_offs(g, cx - 1, cy    , cz + 1);
	offs[ 9] = cgrid_cell_offs(g, cx    , cy    , cz + 1);
	offs[10] = cgrid_cell_offs(g, cx + 1, cy    , cz + 1);
	offs[11] = cgrid_cell_offs(g, cx - 1, cy + 1, cz + 1);
	offs[12] = cgrid_cell_offs(g, cx    , cy + 1, cz + 1);
	offs[13] = cgrid_cell_offs(g, cx + 1, cy + 1, cz + 1);
	
	p = g->data[offs[0]];
	i = 0;
	while (p) {
		// same-cell
		q = p;
		j = i + 1;
		if (j >= q->nid) {
			j = 0;
			q = q->next;
		}
		
		while (q) {
			// DEBUG("%p:%d (%d) vs %p:%d (%d)\n",
			//	p, i, p->id[i], q, j, q->id[j]);
			gb_collision(p->id[i], q->id[j]);
	
			j++;
			if (j == q->nid) {
				j = 0;
				q = q->next;
			}
		}
	
		// cross-cell
		for (c = 1; c < 14; c++)
		if (offs[c] != -1) {
			q = g->data[offs[c]];
			j = 0;
		
			while (q) {
				gb_collision(p->id[i], q->id[j]);

				j++;
				if (j == q->nid) {
					j = 0;
					q = q->next;
				}
			}
		}

		// iterate
		i++;
		if (i == p->nid) {
			i = 0;
			p = p->next;
		}
	}
}

void cgrid_interactions(cgrid *g)
{
	int cx, cy, cz;

	for (cz = 0; cz < g->zcells; cz++) {	
		for (cy = 0; cy < g->ycells; cy++) {
			for (cx = 0; cx < g->xcells; cx++) {
				cgrid_cell_interactions(g, cx, cy, cz);
			}
		}
	}
}

int main(int argc, char **argv, char **envp)
{
	double x, y, z;
	int i;
	cgrid *g;
	
	gb_init("Grid of centers", argc, argv);
	
	g = cgrid_new();
	
	while(1) {
		gb_timer_start(GB_COLLISION_PHASE);
		
		cgrid_reset(g);

		gb_timer_start(GB_POSITION_UPDATE);
		for (i = 0; i < gb_nobj(); i++) {
			gb_obj_coord(i, &x, &y, &z);
			cgrid_add_sphere(g, x, y, z, i);
		}
		
		gb_timer_stop(GB_POSITION_UPDATE);

		gb_tell_nodes(g->nodes);
		gb_tell_memory(
			gb_malloc_total() + g->nodes * sizeof(cgrid_cell));
		
		cgrid_interactions(g);
		
		gb_timer_stop(GB_COLLISION_PHASE);
		
		gb_next_step();
	}

	return(0);
}


