#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <errno.h>
#include "genbench.h"

#define DEBUG(s...)	fprintf(stderr, s)

// *******************************************************************
// Parameters
// *******************************************************************
#define ARG_END		0
#define ARG_BOOL	1
#define ARG_INT		2
#define ARG_DOUBLE	3
#define ARG_STRING	4

static struct cmdline_option_t {
	int type;
	char *id;
	char *modifier;
	char *description;
	char *units;
	double value;
	char *string;
} cmdline_options[] = {
{ ARG_INT, "N_OBJS", "-n", "Number of particles to simulate", "", 10000 },
{ ARG_DOUBLE, "MASS", "-m", "Particle mass", " kg", 1.0 },
{ ARG_DOUBLE, "MASS_DELTA", "-md", "Particle mass delta", " kg", 0.0 },
{ ARG_DOUBLE, "RADIUS", "-r", "Particle potential radius", " m", 0.24 },
{ ARG_DOUBLE, "RADIUS_DELTA", "-rd", "Particle potential delta", " m", 0.0 },
{ ARG_DOUBLE, "DENSITY", "-d", "Particles per cube unit", " m^(-3)", 100.0 },
{ ARG_DOUBLE, "PEAK_FORCE", "-p", "Peak repulsive force", " N", 100.0 },
{ ARG_DOUBLE, "PEAK_FORCE_DELTA", "-pd",
	"Peak repulsive force delta", " N", 0.0 },
{ ARG_DOUBLE, "DRAG_COEFF", "-b",
	"Linear drag force coefficient", " N/(m/s)", 1.0 },
{ ARG_DOUBLE, "GRAVITY", "-g", "Gravitational acceleration", " m/s^2", 9.81 },
{ ARG_DOUBLE, "TIME_STEP", "-dt", "Time step", " s", 0.004 },
{ ARG_DOUBLE, "FORM_FACTOR", "-ff",
	"Initial distribution form factor (x / y = z / y)", "", 0.5 },
{ ARG_DOUBLE, "RESTITUTION", "-cc",
	"Restitution coefficient of plane constraints", "", 0.2  },
{ ARG_DOUBLE, "JITTER", "-cj",
	"Position jitter on plane constraints", "", 0.01  },
{ ARG_DOUBLE, "RELAX_TIME", "-crt",
	"Simulation space relaxation time", " s", 50 * 0.004 },
{ ARG_DOUBLE, "RELAX_FACTOR", "-crx",
	"Simulation space relaxation factor", "", 1.0 },
{ ARG_INT, "RELAX_MASK", "-crm",
	"Simulation space relaxation mask", "", 2 },
{ ARG_DOUBLE, "MAX_STEP_DIST", "-dxmax",
	"Velocity overflow check", " m/s", 1e6 },
{ ARG_BOOL, "MID_PHASE", "-mid",
	"Measure mid-phase time", "", 0  },
{ ARG_BOOL, "HANDLING", "-hand",
	"Measure collision handling time", "", 0 },
{ ARG_STRING, "STATS_FILE", "-os",
	"Output file for statistics", "/stdout", 0.0, "-" },
{ ARG_STRING, "OUTPUT_FILE", "-o",
	"Output file for state", "none", 0.0, NULL },
{ ARG_INT, "STOP_ITER", "-si", "Stop after given iteration", "/never", -1 },
{ ARG_END, "ARG_END", "", "arg_end", "", 0.0 } };

// *******************************************************************
// 
// *******************************************************************
/*
  Rationale on (not) using rdtsc
  
  The call to gettimeofday() implies switching to kernel mode, which is
  quite slow. This adds a significant bias on measure of short timings.
  There is an instruction (rdtsc) that returns a 64-bit timestamp counter
  incrementing at CPU frequency.
  But rdtsc has a few drawbacks: it is not portable to non-x86 architectures,
  and may give inconsistent results due to various CPU features such as
  frequency scaling, hyperthreading, and core switching.

  Update - 2009-12-17

  Now on suited architectures, gettimeofday() does not lead to a system
  call anymore. In most cases, it executes some shared kernel code in user
  mode on behalf of the calling process. See Linux documentation on virtual
  dynamic shared objects (VDSO) for more detailed info.

*/
#ifndef rdtsc
#define rdtsc(low,high) \
     __asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))
#endif


// *******************************************************************
// Internals
// *******************************************************************
#define GB_MAX_TIMERS	32
typedef struct gb_clock_t {
	struct timeval start;
	struct timeval total;
	int n_calls;
} gb_clock;


typedef struct gb_obj_t {
	double x, y, z;
	double vx, vy, vz;
	float radius, inv_mass, peak_force;
	float r, g, b;
} gb_obj;

typedef struct gb_statistics_t {
	// simulation parameters
	int n_objs;
	double x_min, x_max, y_min, y_max, z_min, z_max;
	double drag_coeff, gravity, restitution, jitter;
	double time_step;
	double relax_time;
	double rx_min, rx_max, ry_min, ry_max, rz_min, rz_max;
	double max_step_dist;
	
	double max_radius;
	int uniform_radius;
	int stop_iter, measure_mid, measure_hand;
	
	// simulation state
	char *sim_name;
	int frame;
	
	// per-frame
	int pairs, true_pos;
	int bounces, nodes;
	int velocity_overflows;
	size_t memory;
	
	// output
	FILE *stats_out;
	FILE *state_out;
} gb_statistics;


static gb_clock gb_timer[GB_MAX_TIMERS];
static gb_obj *gb_objs;
static gb_statistics gb_stats;



// *******************************************************************
// Command-line handling
// *******************************************************************
static void help_message()
{
	int i;
	
	fprintf(stdout, "\nBenchmarking broad-phase algorithm \"%s\"\n\n"
		"Command line options [default]:\n\n",
		gb_stats.sim_name);

	i = 0;
	while(cmdline_options[i].type != ARG_END) {
		fprintf(stdout, " %s\t%s", cmdline_options[i].modifier,
			cmdline_options[i].description);
		
		switch(cmdline_options[i].type) {
		case ARG_STRING:
			fprintf(stdout, " [%s%s]\n",
				(cmdline_options[i].string) ?
				cmdline_options[i].string : "",
				cmdline_options[i].units);
			break;
		case ARG_DOUBLE:
			fprintf(stdout, " [%g%s]\n",
				cmdline_options[i].value,
				cmdline_options[i].units);
			break;
		case ARG_INT:
			fprintf(stdout, " [%d%s]\n",
				(int)cmdline_options[i].value,
				cmdline_options[i].units);
			break;
		case ARG_BOOL:
		default:
			fprintf(stdout, " [%s]\n",
				(cmdline_options[i].value == 0.0) ?
				"disabled" : "enabled");
			break;
		}
		i++;
	}
	
	fprintf(stdout, "\n");
}

static int arg_index(char *id, int type)
{
	int i = 0;
	
	while(cmdline_options[i].type != ARG_END) {
		if ((cmdline_options[i].type == type)
		&&  (!strcmp(cmdline_options[i].id, id)))
			return(i);
		i++;
	}

	DEBUG("BUG: arg_index(): Unknown id '%s'\n", id);
	exit(1);
}

static int arg_find_modifier(char *modifier)
{
	int i = 0;

	while(cmdline_options[i].type != ARG_END) {
		if (!strcmp(modifier, cmdline_options[i].modifier))
			return(i);
		i++;
	}
	
	return(i);
}

static double arg_double(char *id)
{
	return(cmdline_options[arg_index(id, ARG_DOUBLE)].value);
}

static int arg_int(char *id)
{
	return(cmdline_options[arg_index(id, ARG_INT)].value);
}

static int arg_bool(char *id)
{
	return(cmdline_options[arg_index(id, ARG_BOOL)].value != 0.0);
}

static char *arg_string(char *id)
{
	return(cmdline_options[arg_index(id, ARG_STRING)].string);
}

static void arg_parse(int argc, char **argv)
{
	int t, i;
	char *e;
	
	t = 1;
	while(t < argc) {
		if ((!strcmp(argv[t], "--help")) || (!strcmp(argv[t], "-h"))) {
			help_message();
			exit(0);
		}
		
		i = arg_find_modifier(argv[t]);
		if (cmdline_options[i].type == ARG_END) {
			help_message();
			exit(1);
		} else if (cmdline_options[i].type == ARG_BOOL) {
			cmdline_options[i].value = 1;
			t++;
		} else if (t + 1 >= argc) {
			DEBUG("'%s' needs an argument\n", argv[t]);
			exit(1);
		} else if (cmdline_options[i].type == ARG_INT) {
			cmdline_options[i].value = strtol(argv[t + 1], &e, 0);
			if ((*e != 0) || (e == argv[t + 1])) {
				DEBUG("'%s': invalid integer argument\n",
					argv[t]);
				exit(1);
			}
			t += 2;
		} else if (cmdline_options[i].type == ARG_DOUBLE) {
			cmdline_options[i].value = strtod(argv[t + 1], &e);
			if ((*e != 0) || (e == argv[t + 1])) {
				DEBUG("'%s': invalid value\n", argv[t]);
				exit(1);
			}
			t += 2;
		} else if (cmdline_options[i].type == ARG_STRING) {
			cmdline_options[i].string = argv[t + 1];
			t += 2;
		}
	}
}

static void arg_print_values(FILE *f)
{
	int i = 0;
	
	while(cmdline_options[i].type != ARG_END) {
		switch(cmdline_options[i].type) {
		case ARG_STRING:
			fprintf(f, "%s \"%s\"",
				cmdline_options[i].id,
				(cmdline_options[i].string) ?
				cmdline_options[i].string : "");
			break;
		case ARG_DOUBLE:
			fprintf(f, "%s %10g",
				cmdline_options[i].id,
				cmdline_options[i].value);
			break;
		case ARG_INT:
			fprintf(f, "%s %10i",
				cmdline_options[i].id,
				(int)cmdline_options[i].value);
			break;
		case ARG_BOOL:
		default:
			fprintf(f, "%s %s",
				cmdline_options[i].id,
				(cmdline_options[i].value == 0.0) ?
				"no" : "yes");
			break;
		}
		
		fprintf(f, "         ");
		i++;
	}
	fprintf(f, "\n");
}

// ***************************************************************************
// Utility random function
// ***************************************************************************
static double drand(double min, double max)
{
	double n01;
	
	n01 = random() / (1.0 + RAND_MAX);
	
	return(min + (n01 * (max - min)));
}

// ***************************************************************************
// Initialization
// ***************************************************************************
void gb_init(char *sim_name, int argc, char **argv)
{
	// parameters
	double mass, mass_delta, obj_mass;
	double radius, radius_delta;
	double density;
	double peak_force, peak_force_delta;
	double form_factor;
	double relax_factor;
	int relax_mask;

	char *filename;
	double dimy, dimxz;
	int t;
	
	// Command line arguments
	gb_stats.sim_name = sim_name;
	arg_parse(argc, argv);
	
	gb_stats.n_objs = arg_int("N_OBJS");
	gb_stats.drag_coeff = arg_double("DRAG_COEFF");
	gb_stats.gravity = arg_double("GRAVITY");
	gb_stats.restitution = arg_double("RESTITUTION");
	gb_stats.jitter = arg_double("JITTER");
	gb_stats.time_step = arg_double("TIME_STEP");
	gb_stats.relax_time = arg_double("RELAX_TIME");
	gb_stats.max_step_dist = arg_double("MAX_STEP_DIST");
	gb_stats.stop_iter = arg_int("STOP_ITER");
	gb_stats.measure_mid = arg_bool("MID_PHASE");
	gb_stats.measure_hand = arg_bool("HANDLING");

	mass = arg_double("MASS");
	mass_delta = arg_double("MASS_DELTA");
	radius = arg_double("RADIUS");
	radius_delta = arg_double("RADIUS_DELTA");
	density = arg_double("DENSITY");
	peak_force = arg_double("PEAK_FORCE");
	peak_force_delta = arg_double("PEAK_FORCE_DELTA");
	form_factor = arg_double("FORM_FACTOR");
	relax_factor = arg_double("RELAX_FACTOR");
	relax_mask = arg_int("RELAX_MASK");

	gb_stats.max_radius = radius + radius_delta;
	gb_stats.uniform_radius = (radius_delta == 0.0);
	
	filename = arg_string("OUTPUT_FILE");
	if (!filename) {
		gb_stats.state_out = NULL;
	} else if (!strcmp(filename, "-")) {
		gb_stats.state_out = stdout;
	} else if (!strcmp(filename, "-o")) {
		gb_stats.state_out = fopen(filename, "w");
		if (!gb_stats.state_out) {
			DEBUG("%s: %s\n", filename, strerror(errno));
			exit(1);
		}
	}

	filename = arg_string("STATS_FILE");
	if (!filename) {
		gb_stats.stats_out = NULL;
	} else if (!strcmp(filename, "-")) {
		gb_stats.stats_out = (gb_stats.state_out == stdout) ?
			stderr : stdout;
	} else {
		gb_stats.stats_out = fopen(filename, "w");
		if (!gb_stats.stats_out) {
			DEBUG("%s: %s\n", filename, strerror(errno));
			exit(1);
		}
	}
	
	
	// Generate objects
	gb_objs = (gb_obj *)malloc(gb_stats.n_objs * sizeof(gb_obj));
	if (!gb_objs) {
		DEBUG("malloc(%d objs) failed\n", gb_stats.n_objs);
		exit(1);
	}
	
	dimy = cbrt((double)gb_stats.n_objs
		/ (density * form_factor * form_factor));
	dimxz = form_factor * dimy;

	gb_stats.x_min = gb_stats.z_min = -0.5 * dimxz;
	gb_stats.x_max = gb_stats.z_max =  0.5 * dimxz;
	gb_stats.y_min = 0.0;
	gb_stats.y_max = dimy;
	
	gb_stats.rx_min = gb_stats.x_min; gb_stats.rx_max = gb_stats.x_max;
	gb_stats.ry_min = gb_stats.y_min; gb_stats.ry_max = gb_stats.y_max;
	gb_stats.rz_min = gb_stats.z_min; gb_stats.rz_max = gb_stats.z_max;
	if (relax_mask &  1) gb_stats.rx_min *= relax_factor;
	if (relax_mask &  2) gb_stats.rx_max *= relax_factor;
	if (relax_mask &  4) gb_stats.ry_min *= relax_factor;
	if (relax_mask &  8) gb_stats.ry_max *= relax_factor;
	if (relax_mask & 16) gb_stats.rz_min *= relax_factor;
	if (relax_mask & 32) gb_stats.rz_max *= relax_factor;
	
	srandom(1);

	for (t = 0; t < gb_stats.n_objs; t++) {
		gb_objs[t].x = drand(gb_stats.x_min, gb_stats.x_max);
		gb_objs[t].y = drand(gb_stats.y_min, gb_stats.y_max);
		gb_objs[t].z = drand(gb_stats.z_min, gb_stats.z_max);

		gb_objs[t].vx = gb_objs[t].vy = gb_objs[t].vz = 0.0;
		
		obj_mass = drand(mass - mass_delta, mass + mass_delta);
		gb_objs[t].inv_mass = 1.0 / obj_mass;
		gb_objs[t].radius = drand(radius - radius_delta,
			radius + radius_delta);
		gb_objs[t].peak_force = drand(peak_force - peak_force_delta,
			peak_force + peak_force_delta);
		
		// color codes
		if (peak_force_delta > 0.0) {
			gb_objs[t].r = 0.5 * (gb_objs[t].peak_force
				- (peak_force - peak_force_delta))
				/ peak_force_delta;
		} else {
			gb_objs[t].r = 0.5;
		}
		
		if (radius_delta > 0.0) {
			gb_objs[t].g = 0.5 * (gb_objs[t].radius
				- (radius - radius_delta)) / radius_delta;
		} else {
			gb_objs[t].g = 0.5;
		}

		if (mass_delta > 0.0) {
			gb_objs[t].b = 0.5 * (obj_mass
				- (mass - mass_delta)) / mass_delta;
		} else {
			gb_objs[t].b = 0.5;
		}
	}
	

	gb_stats.frame = 0;

	gb_stats.pairs = gb_stats.true_pos = 0;
	gb_stats.bounces = 0;
	gb_stats.nodes = 0;
	gb_stats.velocity_overflows = 0;
	gb_stats.memory = 0;
	
	gb_timers_init();

	// Initial distribution statistics
	if (gb_stats.stats_out) {
		fprintf(gb_stats.stats_out, "# Parameters\n");
		arg_print_values(gb_stats.stats_out);
	}

	// Output initial distribution (frame -1)
	gb_dump_state(1);
}

// ***************************************************************************
// Memory stats:
// ***************************************************************************
size_t gb_malloc_total()
{
	return(gb_stats.n_objs * sizeof(gb_obj));
}

// ***************************************************************************
// Accessors
// ***************************************************************************
int  gb_nobj()
{
	return(gb_stats.n_objs);
}

void gb_obj_coord(int n, double *x, double *y, double *z)
{
	// if ((n < 0) || (n >= gb_stats.n_objs)) return;
	
	if (x) *x = gb_objs[n].x;
	if (y) *y = gb_objs[n].y;
	if (z) *z = gb_objs[n].z;
}

void gb_obj_color(int n, double *r, double *g, double *b)
{
	// if ((n < 0) || (n >= gb_stats.n_objs)) return;
	
	if (r) *r = gb_objs[n].r;
	if (g) *g = gb_objs[n].g;
	if (b) *b = gb_objs[n].b;
}

double gb_obj_radius(int n)
{
	// if ((n < 0) || (n >= gb_stats.n_objs)) return(0.0);

	return(gb_objs[n].radius);
}

int gb_uniform_radius()
{
	return(gb_stats.uniform_radius);
}

double gb_max_radius()
{
	return(gb_stats.max_radius);
}

void gb_tell_nodes(int nodes)
{
	gb_stats.nodes = nodes;
}

void gb_tell_memory(size_t mem)
{
	gb_stats.memory = mem;
}

#define GB_MIN(a, b)	(((a) < (b)) ? (a) : (b))
#define GB_MAX(a, b)	(((a) > (b)) ? (a) : (b))
double gb_min_x()
{ return(GB_MIN(gb_stats.x_min, gb_stats.rx_min) - 1.1 * gb_stats.max_radius); }
double gb_max_x()
{ return(GB_MAX(gb_stats.x_max, gb_stats.rx_max) + 1.1 * gb_stats.max_radius); }
double gb_min_y()
{ return(GB_MIN(gb_stats.y_min, gb_stats.ry_min) - 1.1 * gb_stats.max_radius); }
double gb_max_y()
{ return(GB_MAX(gb_stats.y_max, gb_stats.ry_max) + 1.1 * gb_stats.max_radius); }
double gb_min_z()
{ return(GB_MIN(gb_stats.z_min, gb_stats.rz_min) - 1.1 * gb_stats.max_radius); }
double gb_max_z()
{ return(GB_MAX(gb_stats.z_max, gb_stats.rz_max) + 1.1 * gb_stats.max_radius); }

// ***************************************************************************
// Collision narrow-phase handler
// ***************************************************************************
void gb_collision(int n1, int n2)
{
	double dx, dy, dz, sqd, d, corr, mind;
	int has_d;
	
	// if ((n < 0) || (n >= gb_stats.n_objs)) return;

	gb_stats.pairs++;
	
	dx = gb_objs[n2].x - gb_objs[n1].x;
	dy = gb_objs[n2].y - gb_objs[n1].y;
	dz = gb_objs[n2].z - gb_objs[n1].z;
	
	sqd = (dx * dx) + (dy * dy) + (dz * dz);
	d = 0.0;
	has_d = 0;
	
	mind = gb_objs[n1].radius;
	if (sqd < mind * mind) {
		gb_stats.true_pos++;

		if (gb_stats.measure_hand)
			gb_timer_start(GB_COLLISION_HANDLING);
	
		d = sqrt(sqd);
		has_d = 1;
		
		corr = gb_stats.time_step * gb_objs[n1].peak_force
			* gb_objs[n2].inv_mass * ((mind - d) / mind);
	
		// normalize difference vector
		corr /= (d < 1e-7) ? 1e-7 : d;

		gb_objs[n2].vx += corr * dx;
		gb_objs[n2].vy += corr * dy;
		gb_objs[n2].vz += corr * dz;

		if (gb_stats.measure_hand)
			gb_timer_stop(GB_COLLISION_HANDLING);
	}

	mind = gb_objs[n2].radius;
	if (sqd < mind * mind) {
		gb_stats.true_pos++;

		if (gb_stats.measure_hand)
			gb_timer_start(GB_COLLISION_HANDLING);
	
		if (!has_d) d = sqrt(sqd);
		
		corr = gb_stats.time_step * gb_objs[n2].peak_force
			* gb_objs[n1].inv_mass * ((mind - d) / mind);
	
		// normalize difference vector
		corr /= (d < 1e-7) ? 1e-7 : d;

		gb_objs[n1].vx -= corr * dx;
		gb_objs[n1].vy -= corr * dy;
		gb_objs[n1].vz -= corr * dz;

		if (gb_stats.measure_hand)
			gb_timer_stop(GB_COLLISION_HANDLING);
	}
}

// ***************************************************************************
// Frame-to-frame transition
// ***************************************************************************
#define BOUND_SPEED(v) \
	if ((v) * gb_stats.time_step >= gb_stats.max_step_dist) { \
		gb_stats.velocity_overflows++; \
		v = gb_stats.max_step_dist / gb_stats.time_step; \
	} \
	if (-(v) * gb_stats.time_step >= gb_stats.max_step_dist) { \
		gb_stats.velocity_overflows++; \
		v = -gb_stats.max_step_dist / gb_stats.time_step; \
	}

void gb_next_step()
{
	double m, jitter;
	int t;
	
	// sudden constraint relaxation
	if ((gb_stats.relax_time > 0.0)
	&&  (gb_stats.frame * gb_stats.time_step >= gb_stats.relax_time)) {
		gb_stats.relax_time = 0.0;
		
		gb_stats.x_min = gb_stats.rx_min;
		gb_stats.x_max = gb_stats.rx_max;
		gb_stats.y_min = gb_stats.ry_min;
		gb_stats.y_max = gb_stats.ry_max;
		gb_stats.z_min = gb_stats.rz_min;
		gb_stats.z_max = gb_stats.rz_max;
	}
	
	// time integration
	for (t = 0; t < gb_stats.n_objs; t++) {
		// Gravity acceleration (speed update)
		gb_objs[t].vy -= gb_stats.gravity * gb_stats.time_step;

		// Friction
		m = 1.0 - (gb_stats.time_step * gb_objs[t].inv_mass
			* gb_stats.drag_coeff);
		gb_objs[t].vx *= m;
		gb_objs[t].vy *= m;
		gb_objs[t].vz *= m;

		// Velocity bounds
		BOUND_SPEED(gb_objs[t].vx);
		BOUND_SPEED(gb_objs[t].vy);
		BOUND_SPEED(gb_objs[t].vz);
		
		// Position update
		gb_objs[t].x += gb_objs[t].vx * gb_stats.time_step;
		gb_objs[t].y += gb_objs[t].vy * gb_stats.time_step;
		gb_objs[t].z += gb_objs[t].vz * gb_stats.time_step;
		
		// Boundaries
		jitter = drand(-gb_stats.jitter, gb_stats.jitter);
		
		if (gb_objs[t].x >= gb_stats.x_max + jitter) {
			gb_objs[t].x = gb_stats.x_max + jitter;
			gb_objs[t].vx = -gb_objs[t].vx * gb_stats.restitution;
			gb_stats.bounces++;
		}
		if (gb_objs[t].x <= gb_stats.x_min + jitter) {
			gb_objs[t].x = gb_stats.x_min + jitter;
			gb_objs[t].vx = -gb_objs[t].vx * gb_stats.restitution;
			gb_stats.bounces++;
		}
		if (gb_objs[t].y >= gb_stats.y_max + jitter) {
			gb_objs[t].y = gb_stats.y_max + jitter;
			gb_objs[t].vy = -gb_objs[t].vy * gb_stats.restitution;
			gb_stats.bounces++;
		}
		if (gb_objs[t].y <= gb_stats.y_min + jitter) {
			gb_objs[t].y = gb_stats.y_min + jitter;
			gb_objs[t].vy = -gb_objs[t].vy * gb_stats.restitution;
			gb_stats.bounces++;
		}
		if (gb_objs[t].z >= gb_stats.z_max + jitter) {
			gb_objs[t].z = gb_stats.z_max + jitter;
			gb_objs[t].vz = -gb_objs[t].vz * gb_stats.restitution;
			gb_stats.bounces++;
		}
		if (gb_objs[t].z <= gb_stats.z_min + jitter) {
			gb_objs[t].z = gb_stats.z_min + jitter;
			gb_objs[t].vz = -gb_objs[t].vz * gb_stats.restitution;
			gb_stats.bounces++;
		}
	}
	
	// output
	gb_dump_state(0);
	gb_dump_stats(gb_stats.frame == 0);
	
	if (gb_stats.velocity_overflows > 0) {
		DEBUG("Warning: %d velocity overflows detected.\n",
			gb_stats.velocity_overflows);
	}

	// exit
	if (gb_stats.frame == gb_stats.stop_iter) {
		exit(0);
	}
	
	// init next frame
	gb_stats.frame++;

	gb_stats.pairs = gb_stats.true_pos = 0;
	gb_stats.bounces = 0;
	gb_stats.nodes = 0;
	gb_stats.velocity_overflows = 0;

	for (t = 0; t < GB_MAX_TIMERS; t++)
		gb_timer_clear(t);
}

// ***************************************************************************
// Stats Output
// ***************************************************************************
void gb_dump_stats(int with_header)
{
	int t;
	double mx, my, mz, vx, vy, vz, v;
	
	if (!gb_stats.stats_out) return;

	// mean
	mx = my = mz = 0.0;
	for (t = 0; t < gb_stats.n_objs; t++) {
		mx += gb_objs[t].x;
		my += gb_objs[t].y;
		mz += gb_objs[t].z;
	}
	mx /= (double)gb_stats.n_objs;
	my /= (double)gb_stats.n_objs;
	mz /= (double)gb_stats.n_objs;
	
	// std dev
	vx = vy = vz = 0.0;
	for (t = 0; t < gb_stats.n_objs; t++) {
		v = (gb_objs[t].x - mx); vx += v * v;
		v = (gb_objs[t].y - my); vy += v * v;
		v = (gb_objs[t].z - mz); vz += v * v;
	}
	vx = sqrt(vx / (double)gb_stats.n_objs);
	vy = sqrt(vy / (double)gb_stats.n_objs);
	vz = sqrt(vz / (double)gb_stats.n_objs);
	
	if (with_header) {
		fprintf(gb_stats.stats_out,
			"# \"Frame\" \"Objects\" \"mean x\" \"mean y\" "
			"\"mean z\" \"dev x\" \"dev y\" \"dev z\" \"bounces\" "
			"\"Coll.\" \"True+\"  \"Nodes\" \"Coll. Time\" "
			"\"Mid. Ph.\"  \"Handling\" \"Update Time\""
			" \"Memory\"\n");
	}
	
	fprintf(gb_stats.stats_out,
		" %6d %8d %8.3f %8.3f %8.3f %7.3f %7.3f %7.3f %8d",
		gb_stats.frame, gb_stats.n_objs, mx, my, mz, vx, vy, vz,
		gb_stats.bounces);
	
	fprintf(gb_stats.stats_out,
		" %8d %8d %9d %11.3f %11.3f %11.3f %11.3f %11zd\n",
		gb_stats.pairs, gb_stats.true_pos,
		gb_stats.nodes,
		gb_timer_ms(GB_COLLISION_PHASE),
		gb_stats.measure_mid ? gb_timer_ms(GB_MIDDLE_PHASE) : -1.0,
		gb_stats.measure_hand ? gb_timer_ms(GB_COLLISION_HANDLING)
			: -1.0,
		gb_timer_ms(GB_POSITION_UPDATE),
		gb_stats.memory);
	
	fflush(gb_stats.stats_out);
}

// ***************************************************************************
// State Output (this part is not portable)
// ***************************************************************************
void gb_dump_state(int with_header)
{
	int t;
	
	if (!gb_stats.state_out) return;

	if (with_header) {
		t = 0x10203040;
		fwrite(&t, 4, 1, gb_stats.state_out);
		fwrite(&gb_stats.n_objs, 4, 1, gb_stats.state_out);

		for (t = 0; t < gb_stats.n_objs; t++) {
			fwrite(&gb_objs[t].radius, 4, 1, gb_stats.state_out);
			fwrite(&gb_objs[t].r, 4, 1, gb_stats.state_out);
			fwrite(&gb_objs[t].g, 4, 1, gb_stats.state_out);
			fwrite(&gb_objs[t].b, 4, 1, gb_stats.state_out);
		}
	}
	
	for (t = 0; t < gb_stats.n_objs; t++) {
		fwrite(&gb_objs[t].x, 8, 1, gb_stats.state_out);
		fwrite(&gb_objs[t].y, 8, 1, gb_stats.state_out);
		fwrite(&gb_objs[t].z, 8, 1, gb_stats.state_out);
	}
	
	fflush(gb_stats.state_out);
}

// ***************************************************************************
// Replay helper
// ***************************************************************************
int  gb_load_sim_file(FILE *f)
{
	int t;
	
	if (!f) {
		DEBUG("gl_load_sim_file(): null file\n");
		return(-1);
	}

	gb_stats.state_out = f;
	
	if ((fread(&t, 4, 1, gb_stats.state_out) != 1) || (t != 0x10203040)) {
		DEBUG("Simulation file: bad header\n");
		gb_stats.state_out = NULL;
		return(-1);
	}
	
	if (fread(&gb_stats.n_objs, 4, 1, gb_stats.state_out) != 1) {
		DEBUG("Simulation file: %s\n", strerror(errno));
		gb_stats.state_out = NULL;
		return(-1);
	}
	
	gb_objs = (gb_obj *)malloc(gb_stats.n_objs * sizeof(gb_obj));
	if (!gb_objs) {
		DEBUG("malloc(%d objs) failed\n", gb_stats.n_objs);
		gb_stats.state_out = NULL;
		return(-1);
	}
	
	for (t = 0; t < gb_stats.n_objs; t++) {
		if ((fread(&gb_objs[t].radius, 4, 1, gb_stats.state_out) != 1)
		||  (fread(&gb_objs[t].r, 4, 1, gb_stats.state_out) != 1)
		||  (fread(&gb_objs[t].g, 4, 1, gb_stats.state_out) != 1)
		||  (fread(&gb_objs[t].b, 4, 1, gb_stats.state_out) != 1)) {
			DEBUG("gb_load_sim_file(): %s\n", strerror(errno));
			gb_stats.state_out = NULL;
			return(-1);
		}
	}
	
	return(gb_load_frame());
}

int  gb_load_sim(char *file_name)
{
	FILE *f;

	f = fopen(file_name, "r");
	if (!f) {
		DEBUG("%s: %s\n", file_name, strerror(errno));
		return(-1);
	}

	return(gb_load_sim_file(f));
}

int  gb_load_frame()
{
	int t;
	
	if (!gb_stats.state_out) return(-1);
	
	for (t = 0; t < gb_stats.n_objs; t++) {
		if ((fread(&gb_objs[t].x, 8, 1, gb_stats.state_out) != 1)
		||  (fread(&gb_objs[t].y, 8, 1, gb_stats.state_out) != 1)
		||  (fread(&gb_objs[t].z, 8, 1, gb_stats.state_out) != 1)) {
			gb_stats.state_out = NULL;
			return(-1);
		}
	}
	
	return(0);
}

// ***************************************************************************
// Timers
// ***************************************************************************
void gb_timers_init()
{
	int t;
	
	for (t = 0; t < GB_MAX_TIMERS; t++)
		gb_timer_clear(t);
}

void gb_timer_clear(int id)
{
	// if ((id < 0) || (id >= GB_MAX_TIMERS)) return;
	
	gb_timer[id].total.tv_sec = 0;
	gb_timer[id].total.tv_usec = 0;
	gb_timer[id].n_calls = 0;
}


void gb_timer_start(int id)
{
	// if ((id < 0) || (id >= GB_MAX_TIMERS)) return;

	gettimeofday(&gb_timer[id].start, NULL);
}

void gb_timer_stop(int id)
{
	struct timeval now, diff;
	
	// if ((id < 0) || (id >= GB_MAX_TIMERS)) return;

	gettimeofday(&now, NULL);
	timersub(&now, &gb_timer[id].start, &diff);
	timeradd(&gb_timer[id].total, &diff, &gb_timer[id].total);
	gb_timer[id].n_calls++;
}

double gb_timer_ms(int id)
{
	// if ((id < 0) || (id >= GB_MAX_TIMERS)) return;

	return(1000.0 * gb_timer[id].total.tv_sec
		+ 0.001 * gb_timer[id].total.tv_usec);
}




